from blackjack.deck import Deck
from blackjack.player import Dealer, Player
from blackjack.game import Game


def main(play_again):
    deck = Deck()
    dealer = Dealer()
    player = Player()

    game = Game(deck, dealer, player)
    new_game = game.start(play_again)
    if new_game:
        return main(new_game)
    print("Hope to see you again soon!")


if __name__ == "__main__":
    main(False)


from blackjack.game_setup import BLACKJACK


class Action:

    def __init__(self, deck, person):
        self.deck = deck
        self.person = person

    def _set_ace(self, card_value):
        one, eleven = card_value
        if self.person.score + eleven > BLACKJACK:
            _card_value = one
        else:
            _card_value = eleven
        return _card_value

    def hit(self):
        card = self.deck.cards.pop()
        _card_value = card.card_value()
        if isinstance(_card_value, tuple):
            _card_value = self._set_ace(_card_value)
        self.person.score = _card_value
        self.person.hand.append(card)


from time import sleep

from blackjack.action import Action
from blackjack.game_setup import (
    YES,
    HIT,
    STAY,
    BLACKJACK,
    PRIZE,
    GAME_DELAY,
    DEALER_SCORE_MIN,
    NO,
    PLAYER,
    DEALER,
)
from blackjack.exceptions import InvalidResponseException


class Game:

    def __init__(self, deck, dealer, player):
        self.deck = deck
        self.dealer = dealer
        self.player = player
        self.action = Action

    def start(self, play_again=False):
        while True:
            if play_again:
                sleep(GAME_DELAY)
                print("\nWELCOME BACK! Let's play another game!")               
                start_game = YES[0]
            else:
                start_game = input("Welcome to Blackjack! Would you like to play a game? [yes/no] ").upper()

            if start_game in YES:
                self.place_bet()
                return self.new_game()
            else:
                break

    def place_bet(self):
        try:
            self.player.bet = input("How much money would you like to bet? ")
        except Exception as err:
            print(f"ERROR - {err}\n")
            self.place_bet()
        else:
            self._first_round()

    def _deal_card_message(self, person, time):
        print(f"Dealing {person.kind} Card...")
        sleep(time)

    def _result_message(self, first_round=False, dealer_hand=None):
        print(f"\n{'Result after Round 1' if first_round else 'Final Result'}")
        print("---------------")
        print(self.player)
        print(f"{dealer_hand if first_round else self.dealer}\n")

    def _first_round(self):
        self._deal_card_message(self.player, GAME_DELAY)
        self.player.deal_card(self.deck, self.action)

        self._deal_card_message(self.dealer, GAME_DELAY)
        self.dealer.deal_card(self.deck, self.action)
        dealer_first_card = self.dealer.__str__()

        self._deal_card_message(self.player, GAME_DELAY)
        self.player.deal_card(self.deck, self.action)

        self._deal_card_message(self.dealer, GAME_DELAY)
        self.dealer.deal_card(self.deck, self.action)

        self._result_message(True, dealer_first_card)
        self._process_blackjack(self.player)

    def _process_blackjack(self, person):
        if person.kind == PLAYER:
            if person.score == BLACKJACK:
                self._result_message()
                print(f"Congratulations! You scored Blackjack and win ${self.player.bet * PRIZE}!")            
            elif person.score > BLACKJACK:
                self._result_message()
                print("BUST! House wins!")
            else:
                self._second_round()

        if person.kind == DEALER:
            if person.score == BLACKJACK:
                self._result_message()
                print("House scored Blackjack! You lose!")
            elif person.score > BLACKJACK:
                self._result_message()
                print(f"BUST! Congratulations! You win ${self.player.bet * PRIZE}!")
            else:
                self._compare_hands()

    def _player_action_hit(self):
        self._deal_card_message(self.player, GAME_DELAY)
        self.player.deal_card(self.deck, self.action)
        print(self.player)

    def _player_action_stay(self):
        if self.dealer.score > self.player.score:
            return
        if self.dealer.score < DEALER_SCORE_MIN:
            self._increase_dealer_score()

    def _increase_dealer_score(self):
        self._deal_card_message(self.dealer, GAME_DELAY)
        self.dealer.deal_card(self.deck, self.action)
        print(self.dealer)
        self._player_action_stay()

    def _second_round(self):
        try:
            player_action = input("Would you like another card? [hit/stay] ").upper()
            if player_action not in HIT + STAY:
                raise InvalidResponseException
        except InvalidResponseException as err:
            print(f"{err}\n")
            self._second_round()
        else:
            if player_action in HIT:
                self._player_action_hit()
                self._process_blackjack(self.player)
            else:
                self._player_action_stay()
                self._process_blackjack(self.dealer)
    
    def _compare_hands(self):
        self._result_message()
        if self.player.score > self.dealer.score:
            print(f"Congratulations! You win ${self.player.bet * PRIZE}!")
        elif self.player.score == self.dealer.score:
            print("Keep your money. We have a Tie!")
        else:
            print("Bummer! House wins!")
        
    def new_game(self):
        sleep(GAME_DELAY)
        try:
            next_game = input("\nThanks for playing! Would you like to play another game? [yes/no] ").upper()
            if next_game not in YES + NO:
                raise InvalidResponseException
        except InvalidResponseException as err:
            print(f"{err}\n")
            self.new_game()
        else:
            return next_game in YES


from blackjack.exceptions import InvalidBetException, InvalidScoreException

class Player:

    def __init__(self):
        self._score = 0
        self._bet = 0
        self.hand = []
        self.kind = self.__class__.__name__

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, value):
        if value < 0:
            raise InvalidScoreException
        self._score += value
        return self._score

    @property
    def bet(self):
        return self._bet

    @bet.setter
    def bet(self, value):
        value = int(value)
        if value < 0:
            raise InvalidBetException
        self._bet = value
        return self._bet

    def deal_card(self, deck, action):
        action(deck, self).hit()

    def __str__(self):
        stats = (
            f"{self.__class__.__name__} Bet: {self.bet}\n"
            f"{self.__class__.__name__} Score: {self.score}\n"
            f"{self.__class__.__name__} Hand: {', '.join(f'({card.suit.title()}, {card.rank})' for card in self.hand)}"
        )
        return stats


class Dealer(Player):

    def __init__(self):
        Player.__init__(self)

    def __str__(self):
        stats = (
            f"{self.__class__.__name__} Score: {self.score}\n"
            f"{self.__class__.__name__} Hand: {', '.join(f'({card.suit.title()}, {card.rank})' for card in self.hand)}"
        )
        return stats

    @property
    def bet(self):
        raise AttributeError("Dealer cannot place a bet")


from dataclasses import dataclass, field
from random import shuffle


@dataclass(frozen=True)
class Card:
    suit: str
    rank: str

    def card_value(self):
        if self.rank in "J Q K".split():
            return 10
        elif self.rank == "A":
            return (1, 11)
        else:
            return int(self.rank)

    def __str__(self):
        return f"{self.__class__.__name__}({self.suit}, {self.rank})"


@dataclass
class Deck:

    cards: list = field(default_factory=list)   

    def __post_init__(self):
        suits = "clubs diamonds hearts spades".split()
        _cards = "2 3 4 5 6 7 8 9 10 J Q K A".split()
        deck = []
        for suit in suits:
            for card in _cards:
                deck.append(Card(suit, card))
        shuffle(deck)
        self.cards = deck

    def __str__(self):        
        return f"{self.__class__.__name__}({', '.join(card.__str__() for card in self.cards)})"


class InvalidBetException(Exception):

    def __init__(self):
        self.message = "please provide a positive integer value"
        super().__init__(self.message)


class InvalidScoreException(Exception):

    def __init__(self):
        self.message = "value should be greater than 0"
        super().__init__(self.message)


class InvalidResponseException(Exception):

    def __init__(self):
        self.message = "please enter an acceptable value"
        super().__init__(self.message)